//TODO Is formula for std-deviation -> auto-scale correct (strange *2 ?)
/**************************************************************
 This file is part of
          Shock Filter for AviSynth by Wolfgang Boiger
 Copyright (C)  2006-2008  Wolfgang Boiger, Berlin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2, as published by the Free Software Foundation in
 June 1991 (see 'GNUGPLv2.txt').
 
 There is no guaranty that the rights of third parties
 (including software patents) are not infringed upon by the
 execution or distribution of this program.
 Referring to section 8 of the General Public License, any use
 of this program (including all activities mentioned in
 section 0 and including execution) is restricted to countries
 where any of these activities do not infringe upon any such
 rights of third parties. It lies with the user to verify the
 compliance of his or her use (especially concerning but not
 limited to distribution) of this program with the rights of
 any third parties in the respective countries or regions.
 
 This program is distributed in the hope that it will be
 useful, but WITHOUT ANY WARRANTY; without even the implied
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public
 License 2 along with this program; if not, write to the
   Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **************************************************************/



#define  VersionString  "2008.02.23"
#define  LegalInfoString  "* Shock Filter for AviSynth by Wolfgang Boiger *\nVersion " VersionString " ;  built  " __DATE__ "  " __TIME__ " .\nCopyright (C)  2006-2008  Wolfgang Boiger, Berlin.\nLicensed under GNU GPL version 2 with possible geographic restrictions.\nThis software comes without any warranty, even without warranty\nof merchantability or fitness for a particular purpose.\nSee files 'AVSShock.htm' and 'GNUGPLv2.txt' for details."

#define  ErosionDilation3x3  1  /* Which pixels are taken into account for ero/dil?  1:9-pixel-block  0:5-pixel-stencil */
#define  BlurStandardDeviation  1  /* 1:blur-intensity=standard-deviation of Gauss-Kernel  0:blur-intensity=scale of autonomous heat equation u_t=tr(u_xx) */
#define  BlurIntensityDefault  1.0  /* Default Blurring intensity (treated according to BlurStandardDeviation */



#include  <stdlib.h>
#include  <math.h>
#include  <float.h>

#include  "AviSynth_C.h"



/* ShockFilterDataStruct holds data needed by the filter GetFrame function to decide what to do */
typedef  struct
{
  int  Steps;  /* Number of iteration steps of the shock filter */
  double  Intensity;  /* Size of one iteration step. Resulting scale is Steps*Intensity */
  int  Canny;  /* If !=0, the Haralick-Canny edge detector is used (better and more expensive) */
  int  BlurSize;  /* size of the (one dimensional gaussian) blurring kernel (actual size is 2*BlurSize+1). 0 means no blurring*/
  signed int * xIndex, * yIndex, * xIndexVector, * yIndexVector;  /* Index arrays used to address pixels outside of the image (only needed for convolution) */
  double * BlurKernel, * BlurKernelVector;  /* 1D blurring kernel (of length 2*BlurSize+1), BK points to center pixel */
}
ShockFilterDataStruct;



/*****  Function declaration  *****/

/** AviSynth stuff **/
AVS_Value  AVSC_CC  Shock_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data);
AVS_VideoFrame *  AVSC_CC  Shock_GetFrame(AVS_FilterInfo *  FilterInfo,  int  FrameNo);
void  AVSC_CC  Shock_Free(AVS_FilterInfo *  FilterInfo);

/** Own functions **/
signed int * CreateMirroredIndex(signed int  Index[],  int  Length,  int  Overhang);  /* Will created an index array, which is used to address pixels outside the image. The image appears mirrored at the borders. Returns a pointer to the index of pixel 0 (NOT the first entry in the array) */
double * CreateGaussKernel(double  Kernel[],  int  Size,  double  Intensity);  /* Creates a one dimensional gauss kernel and returns a pointer to its center */
void  DoubleConvolutionByte(float  DestData[],  const  BYTE  SrcData[],  float  TempData[],  int  Width,  int  Height,  int  SrcStride,  int  SrcPitch,  int  KernelSize,  double  Kernel[],  signed int  xIndex[],  signed int  yIndex[]);  /* Blurs a 2d image with the 1d Gauss kernel given and converts from byte to float */
void  ShockEdgeDetectYUY2(signed char  EdgePickData[],  const  BYTE  Src[],  float *  Blur1Data,  float *  Blur2Data,  int  Width,  int  Height,  int  SrcPitch,  int  BlurSize,  double *  BlurKernel,  signed int *  xIndex,  signed int *  yIndex,  int  Canny);
void  ShockEdgeDetectYV12(signed char  EdgePickData[],  const  BYTE  Src[],  float *  Blur1Data,  float *  Blur2Data,  int  Width,  int  Height,  int  SrcPitch,  int  BlurSize,  double *  BlurKernel,  signed int *  xIndex,  signed int *  yIndex,  int  Canny);
void  ShockTransportYUY2(BYTE  Dest[],  const  BYTE  Src[],  signed char  EdgePickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch,  double  Intensity);
void  ShockTransportYV12(BYTE  DestY[],  BYTE  DestU[],  BYTE  DestV[],  const  BYTE  SrcY[],  const  BYTE  SrcU[],  const  BYTE  SrcV[],  signed char  EdgePickData[],  int  Width,  int  Height,  int  DestPitchY,  int  DestPitchU,  int  DestPitchV,  int  SrcPitchY,  int  SrcPitchU,  int  SrcPitchV,  double  Intensity);

/** Colorspace dependent functions implemented in AVSShockCSdep.c **/
void  LineConvolutionByte (float  DestData[],  const  BYTE   SrcData[],  int  DestStride,  int  SrcStride,  signed int  SrcIndex[],  int  Length,  double  Kernel[],  int  KernelSize);
void  LineConvolutionFloat(float  DestData[],  const  float  SrcData[],  int  DestStride,  int  SrcStride,  signed int  SrcIndex[],  int  Length,  double  Kernel[],  int  KernelSize);
void  EdgeDetectorYUY2      (signed char  EdgeSign[],  const  BYTE   SrcLuma[],  int  SrcPitch,  int  Width,  int  Height);
void  EdgeDetectorYV12      (signed char  EdgeSign[],  const  BYTE   SrcLuma[],  int  SrcPitch,  int  Width,  int  Height);
void  EdgeDetectorFloat     (signed char  EdgeSign[],  const  float  SrcLuma[],  int  SrcPitch,  int  Width,  int  Height);
void  EdgeDetectorYUY2Canny (signed char  EdgeSign[],  const  BYTE   SrcLuma[],  int  SrcPitch,  int  Width,  int  Height);
void  EdgeDetectorYV12Canny (signed char  EdgeSign[],  const  BYTE   SrcLuma[],  int  SrcPitch,  int  Width,  int  Height);
void  EdgeDetectorFloatCanny(signed char  EdgeSign[],  const  float  SrcLuma[],  int  SrcPitch,  int  Width,  int  Height);
void  PickLumaYUY2         (BYTE  DestLuma[],  signed char  DestPick[],  const  BYTE  SrcLuma[],  signed char  EdgeSign[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height);
void  PickLumaYV12         (BYTE  DestLuma[],  signed char  DestPick[],  const  BYTE  SrcLuma[],  signed char  EdgeSign[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height);
void  PickLumaYUY2Intensity(BYTE  DestLuma[],  signed char  DestPick[],  const  BYTE  SrcLuma[],  signed char  EdgeSign[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height,  double  Intensity);
void  PickLumaYV12Intensity(BYTE  DestLuma[],  signed char  DestPick[],  const  BYTE  SrcLuma[],  signed char  EdgeSign[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height,  double  Intensity);
void  PickChromaYUY2         (BYTE  DestChroma[],  const  BYTE  SrcChroma[],  signed char  Pick[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height);
void  PickChromaYV12         (BYTE  DestChroma[],  const  BYTE  SrcChroma[],  signed char  Pick[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height);
void  PickChromaYUY2Intensity(BYTE  DestChroma[],  const  BYTE  SrcChroma[],  signed char  Pick[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height,  double  Intensity);
void  PickChromaYV12Intensity(BYTE  DestChroma[],  const  BYTE  SrcChroma[],  signed char  Pick[],  int  DestPitch,  int  SrcPitch,  int  Width,  int  Height,  double  Intensity);



/*****  Function implementation  *****/


AVSC_EXPORT  const char *  AVSC_CC  avisynth_c_plugin_init(AVS_ScriptEnvironment *  Env)
{
  avs_add_function(Env,"Shock","c[Steps]i[Strength]f[Canny]b[Blur]f[BlurSize]i",Shock_Create,0);
  return  "Shock Filter";
}



AVS_Value  AVSC_CC  Shock_Create(AVS_ScriptEnvironment *  Env,  AVS_Value  Args,  void *  Data)
{
  AVS_Value  SrcClipValue, StepsValue, IntensityValue, CannyValue, BlurValue, BlurSizeValue;
  AVS_Value  DestClipValue;
  char * ErrorText;
  int  Steps, BlurSize;
  double  Intensity, BlurIntensity;
  AVS_FilterInfo * FilterInfo;
  AVS_VideoInfo * VideoInfo;
  AVS_Clip * DestClip;
  ShockFilterDataStruct * ShockData;


  /* Store arguments */
  SrcClipValue = avs_array_elt(Args,0);  /* Source clip */
  StepsValue = avs_array_elt(Args,1);  /* Number of steps */
  IntensityValue = avs_array_elt(Args,2);  /* Step size*/
  CannyValue = avs_array_elt(Args,3);  /* Canny on/off */
  BlurValue = avs_array_elt(Args,4);  /* Blur intensity */
  BlurSizeValue = avs_array_elt(Args,5);  /* Size of gaussian kernel */
  
  /* Gather input values, assign defaults, check bounds */
  Steps = (avs_is_int(StepsValue)?avs_as_int(StepsValue):1);  /* (Steps default value) */
  if (Steps<0)  return  avs_new_value_error("Shock: Number of Steps may not be negative\n\n" LegalInfoString "\n");
  Intensity = (avs_is_float(IntensityValue)?((double)(avs_as_float(IntensityValue))):1.0);  /* (Intensity default value) */
  if (Intensity<0.0)  return  avs_new_value_error("Shock: Strength may not be negative");
  if (Intensity>1.0)  return  avs_new_value_error("Shock: Strength may not be higher than 1.0");
  BlurIntensity = (avs_is_float(BlurValue)?((double)(avs_as_float(BlurValue))):BlurIntensityDefault);  /* (BlurIntensity default value) */
  if (BlurIntensity<0.0)  return  avs_new_value_error("Shock: Blur strength may not be negative");
  #if (!(defined(BlurStandardDeviation) && (BlurStandardDeviation!=0)))
  BlurIntensity = sqrt(2*BlurIntensity);
  #endif
  BlurSize = avs_is_int(BlurSizeValue)?avs_as_int(BlurSizeValue):ceil(2*BlurIntensity);  /* (BlurSize default value) */
  if (BlurIntensity<=sqrt(4*sqrt(256.0*DBL_EPSILON/2.0*4.0*atan(1.0)*exp(1.0))))  BlurSize = 0;  /* No blurring at all, this means BlurSize==0 ! */
  if (BlurSize<0)  return  avs_new_value_error("Shock: BlurSize may not be negative");
  
  /* Check properties of input clip */
  DestClip = avs_new_c_filter(Env,&FilterInfo,SrcClipValue,1);
  if (!DestClip)  return  avs_new_value_error("Could not create filter");
  VideoInfo = &FilterInfo->vi;
  /* If clip is only one pixel or no video at all, or if Intensity(=Stepsize) vanishes, we return the clip unchanged (no error in this case) */
  if (!avs_has_video(VideoInfo) || (VideoInfo->width==0) || (VideoInfo->height==0) || ((VideoInfo->width==1) && (VideoInfo->height==1)) || (Steps==0) || (Intensity==0.0))
  {
    DestClipValue = avs_new_value_clip(DestClip);
    avs_release_clip(DestClip);
    if (!avs_defined(DestClipValue))  return  avs_new_value_error("Shock: Could not create new video clip");
    return  DestClipValue;
  }

  ShockData = 0;
  ErrorText = 0;
  /* From now on, when an error is detected, a message will be stored in ErrorText. Hence (ErrorText) indicates that something is wrong. */
  if (!(avs_is_yuy2(VideoInfo) || avs_is_yv12(VideoInfo)))  ErrorText = "Shock: Only YUV color spaces (YUY2 or YV12) allowed";  /* We only allow YUV color spaces, no RGB */
  if (!ErrorText)  ShockData = (ShockFilterDataStruct*)malloc(sizeof(ShockFilterDataStruct));  /* Allocate my data structure */
  if (!ErrorText && !ShockData)  ErrorText = "Shock: Could not allocate memory";
  if (!ErrorText)
  {
    ShockData->Steps = Steps;  /* Store numeric values */
    ShockData->Intensity = Intensity;
    ShockData->BlurSize = BlurSize;
    ShockData->Canny = (avs_is_bool(CannyValue)?avs_as_bool(CannyValue):0);  /* Canny default value */
    if (BlurSize)  /* Allocate memory for blurring, if necessary */
    {
      ShockData->xIndexVector = (int*)malloc(sizeof(int)*(VideoInfo->width+2*BlurSize));
      ShockData->yIndexVector = (int*)malloc(sizeof(int)*(VideoInfo->height+2*BlurSize));
      ShockData->BlurKernelVector = (double*)malloc(sizeof(double)*(2*BlurSize+1));
      if (!(ShockData->xIndexVector && ShockData->yIndexVector && ShockData->BlurKernelVector))  ErrorText = "Shock: Could not allocate memory";
    }
    else
    {
      ShockData->xIndex = ShockData->xIndexVector = 0;
      ShockData->yIndex = ShockData->yIndexVector = 0;
      ShockData->BlurKernel = ShockData->BlurKernelVector = 0;
    }
  }
  if (!ErrorText)
  {
    if (ShockData->xIndexVector)  ShockData->xIndex = CreateMirroredIndex(ShockData->xIndexVector,VideoInfo->width,BlurSize);  /* Fill index vectors */
    if (ShockData->yIndexVector)  ShockData->yIndex = CreateMirroredIndex(ShockData->yIndexVector,VideoInfo->height,BlurSize);
    if (ShockData->BlurKernelVector)  ShockData->BlurKernel = CreateGaussKernel(ShockData->BlurKernelVector,BlurSize,BlurIntensity);  /* Create Gauss kernel if necessary */
    FilterInfo->user_data = ShockData;  /* Structure will be available to (and needed by) GetFrame function */
    FilterInfo->get_frame = Shock_GetFrame;  /* Change frame-computation-function of new clip */
    FilterInfo->free_filter = Shock_Free;  /* Will be called, when filter is destroyed, frees memory occupied by ShockData */
    DestClipValue = avs_new_value_clip(DestClip);  /* Create return value for this function */
    if (!avs_defined(DestClipValue))  ErrorText = "Shock: Could not create new video clip";
  }
  /* No error up to this point => We're done! */
  avs_release_clip(DestClip);
  if (!ErrorText)  return  DestClipValue;
  /* Error occured: Clean up memory, return error message */
  // TODO Free release clip value
  Shock_Free(FilterInfo);
  return  avs_new_value_error(ErrorText);
}



AVS_VideoFrame *  AVSC_CC  Shock_GetFrame(AVS_FilterInfo *  FilterInfo,  int  FrameNo)
{
  /* Shock data */
  int  Steps, BlurSize, Canny;
  double  Intensity;
  signed int * xIndex, * yIndex;
  double * BlurKernel;
  /* Video frames */
  int  Width, Height;
  AVS_VideoFrame * Frame[2];
  BYTE * DataY[2], * DataU[2], * DataV[2];
  #define  Data  DataY
  int  PitchY[2], PitchU[2], PitchV[2];
  #define  Pitch  PitchY
  /* Temporary memory */
  signed char * EdgePickData;
  float * Blur1Data, * Blur2Data;
  char * ErrorText;
  int  k;


  Steps = ((ShockFilterDataStruct *)FilterInfo->user_data)->Steps;
  Intensity = ((ShockFilterDataStruct *)FilterInfo->user_data)->Intensity;
  Canny = ((ShockFilterDataStruct *)FilterInfo->user_data)->Canny;
  BlurSize = ((ShockFilterDataStruct *)FilterInfo->user_data)->BlurSize;
  xIndex = ((ShockFilterDataStruct *)FilterInfo->user_data)->xIndex;
  yIndex = ((ShockFilterDataStruct *)FilterInfo->user_data)->yIndex;
  BlurKernel = ((ShockFilterDataStruct *)FilterInfo->user_data)->BlurKernel;
  
  Width = FilterInfo->vi.width;
  Height = FilterInfo->vi.height;
  
  /* As above, ErrorText will store an error text, if an error occurs */
  ErrorText = 0;
  Frame[1] = 0;
  EdgePickData = 0;
  Blur1Data = Blur2Data = 0;
  Frame[0] = avs_get_frame(FilterInfo->child,FrameNo);
  if (!Frame[0])  ErrorText = "Shock: Could not get source frame";
  if (!ErrorText)
  {
    /* EdgePickFrame will store the dilation/erosion decision for every pixel */
    EdgePickData = (signed char*)malloc(sizeof(signed char)*Width*Height);
    if (!EdgePickData)  ErrorText = "Shock: Could not allocate temporary memory";
  }
  if (!ErrorText&&BlurSize)
  {
    /* Allocate temporary memory for blurring */
    Blur1Data = (float*)malloc(sizeof(float)*Width*Height);
    Blur2Data = (float*)malloc(sizeof(float)*Width*Height);
    if (!(Blur1Data&&Blur2Data))  ErrorText = "Shock: Could not allocate temporary memory";
  }
  /* We make the source frame writable and allocate another frame - then each iteration transfers data between those frames */
  if (!ErrorText)
  {
    avs_make_writable(FilterInfo->env,&(Frame[0]));
    if (!Frame[0])  ErrorText = "Shock: Could make source frame writable";
  }
  if (!ErrorText)
  {
    Frame[1] = avs_new_video_frame(FilterInfo->env,&FilterInfo->vi);
    if (!Frame[0])  ErrorText = "Shock: Could not get second frame";
  }
  if (!ErrorText)
  {
    /* Depending on the color space, we gather the data pointers */
    switch (FilterInfo->vi.pixel_type)
    {
      case AVS_CS_YV12:
        for (k=0 ; k<2 ; k++)
        {
          DataY[k] = avs_get_write_ptr_p(Frame[k],AVS_PLANAR_Y);
          DataU[k] = avs_get_write_ptr_p(Frame[k],AVS_PLANAR_U);
          DataV[k] = avs_get_write_ptr_p(Frame[k],AVS_PLANAR_V);
          PitchY[k] = avs_get_pitch_p(Frame[k],AVS_PLANAR_Y);
          PitchU[k] = avs_get_pitch_p(Frame[k],AVS_PLANAR_U);
          PitchV[k] = avs_get_pitch_p(Frame[k],AVS_PLANAR_V);
          if (!DataY[k] || !DataU[k] || !DataV[k])  ErrorText = "Shock: Could not get data pointers";
        }
      break;
      case AVS_CS_YUY2:
      case AVS_CS_I420:
        for (k=0 ; k<2 ; k++)
        {
          Data[k] = avs_get_write_ptr(Frame[k]);
          Pitch[k] = avs_get_pitch(Frame[k]);
          if (!Data[k])  ErrorText = "Shock: Could not get data pointers";
        }
      break;
      default:
        /*Data[0] = */ DataY[0] = DataU[0] = DataV[0] = 0;
        /*Data[1] = */ DataY[1] = DataU[1] = DataV[1] = 0;
        ErrorText = "Shock internal error: Color space unknown / not allowed";  /* Uhou ?? */
    }
  }
  if (!ErrorText)
  {
    /** Here the core functions are called **/
    switch (FilterInfo->vi.pixel_type)
    {
      case AVS_CS_YUY2:
        for (k=0 ; k<Steps ; k++)
        {
          /* Source frame = Frame[k%2]    target frame = Frame[(k+1)%2] */
          ShockEdgeDetectYUY2(EdgePickData,Data[k%2],Blur1Data,Blur2Data,Width,Height,Pitch[k%2],BlurSize,BlurKernel,xIndex,yIndex,Canny);
          ShockTransportYUY2(Data[(k+1)%2],Data[k%2],EdgePickData,Width,Height,Pitch[(k+1)%2],Pitch[k%2],Intensity);
        }
      break;
      case AVS_CS_YV12:
      case AVS_CS_I420:
        for (k=0 ; k<Steps ; k++)
        {
          /* Source frame = Frame[k%2]    target frame = Frame[(k+1)%2] */
          ShockEdgeDetectYV12(EdgePickData,DataY[k%2],Blur1Data,Blur2Data,Width,Height,PitchY[k%2],BlurSize,BlurKernel,xIndex,yIndex,Canny);
          ShockTransportYV12(DataY[(k+1)%2],DataU[(k+1)%2],DataV[(k+1)%2],DataY[k%2],DataU[k%2],DataV[k%2],EdgePickData,Width,Height,PitchY[(k+1)%2],PitchU[(k+1)%2],PitchV[(k+1)%2],PitchY[k%2],PitchU[k%2],PitchV[k%2],Intensity);
        }
      break;
    }
  }
  /* Free everything (whether error or not) */
  if (Blur2Data)  free(Blur2Data);
  if (Blur1Data)  free(Blur1Data);
  if (EdgePickData)  free(EdgePickData);
  if (Frame[(Steps+1)%2])  avs_release_video_frame(Frame[(Steps+1)%2]);
  if (ErrorText)
  {
    /* If error the final target frame is also released */
    if (Frame[Steps%2])  avs_release_video_frame(Frame[Steps%2]);
    Frame[Steps%2] = 0;
    FilterInfo->error = ErrorText;
  }
  return  Frame[Steps%2];

  #undef  Data
  #undef  Pitch
}



void  AVSC_CC  Shock_Free(AVS_FilterInfo *  FilterInfo)
{
  ShockFilterDataStruct * ShockData;


  ShockData = (ShockFilterDataStruct *)FilterInfo->user_data;
  if (ShockData)
  {
    if (ShockData->BlurKernelVector)  free(ShockData->BlurKernelVector);
    if (ShockData->yIndexVector)  free(ShockData->yIndexVector);
    if (ShockData->xIndexVector)  free(ShockData->xIndexVector);
    free(ShockData);
  }
  FilterInfo->user_data = 0;
}



signed int * CreateMirroredIndex(signed int  Index[],  int  Length,  int  Overhang)
{
  signed int  k;


  Index += Overhang;
  for (k=0 ; k<Length ; k++)  Index[k] = k;
  if (Overhang<Length)
  {
    for (k=0 ; k<Overhang ; k++)  Index[Length+k] = Length-k-1;
    for (k=0 ; k<Overhang ; k++)  Index[-k-1] = k;
  }
  else
  {
    for (k=0 ; k<Length ; k++)  Index[k+Length] = Length-k-1;
    for (k=Length ; k<Overhang ; k++)  Index[k+Length] = Index[k-Length];
    for (k=0 ; k<Overhang ; k++)  Index[-k-1] = Index[k];
  }
  return  Index;
}



double * CreateGaussKernel(double  Kernel[],  int  Size,  double  Intensity)
{
  signed int  k;
  double  Sum;


  Intensity = -1.0/(2.0*Intensity*Intensity);
  Kernel += Size;
  Sum = 0;
  for (k=-Size ; k<0 ; k++)  Sum += Kernel[k] = exp(Intensity*k*k);
  Sum *= 2;
  Sum += Kernel[0] = 1;
  for (k=0 ; k<=Size ; k++)  Kernel[k] = Kernel[-k] /= Sum;
  return  Kernel;
}



void  DoubleConvolutionByte(float  DestData[],  const  BYTE  SrcData[],  float  TempData[],  int  Width,  int  Height,  int  SrcStride,  int  SrcPitch,  int  KernelSize,  double  Kernel[],  signed int  xIndex[],  signed int  yIndex[])
{
  int  k;


  for (k=0 ; k<Width ; k++)  LineConvolutionByte(TempData+k,SrcData+k*SrcStride,Width,SrcPitch,yIndex,Height,Kernel,KernelSize);
  for (k=0 ; k<Height ; k++)  LineConvolutionFloat(DestData+k*Width,TempData+k*Width,1,1,xIndex,Width,Kernel,KernelSize);
}



void  ShockEdgeDetectYUY2(signed char  EdgePickData[],  const  BYTE  Src[],  float *  Blur1Data,  float *  Blur2Data,  int  Width,  int  Height,  int  SrcPitch,  int  BlurSize,  double *  BlurKernel,  signed int *  xIndex,  signed int *  yIndex,  int  Canny)
{
  if (BlurSize)
  {
    DoubleConvolutionByte(Blur2Data,Src,Blur1Data,Width,Height,2,SrcPitch,BlurSize,BlurKernel,xIndex,yIndex);
    if (!Canny)  EdgeDetectorFloat(EdgePickData,Blur2Data,Width,Width,Height);
     else  EdgeDetectorFloatCanny(EdgePickData,Blur2Data,Width,Width,Height);
  }
  else
  {
    if (!Canny)  EdgeDetectorYUY2(EdgePickData,Src,SrcPitch,Width,Height);
     else  EdgeDetectorYUY2Canny(EdgePickData,Src,SrcPitch,Width,Height);
  }
}

   

void  ShockEdgeDetectYV12(signed char  EdgePickData[],  const  BYTE  Src[],  float *  Blur1Data,  float *  Blur2Data,  int  Width,  int  Height,  int  SrcPitch,  int  BlurSize,  double *  BlurKernel,  signed int *  xIndex,  signed int *  yIndex,  int  Canny)
{
  if (BlurSize)
  {
    DoubleConvolutionByte(Blur2Data,Src,Blur1Data,Width,Height,1,SrcPitch,BlurSize,BlurKernel,xIndex,yIndex);
    if (!Canny)  EdgeDetectorFloat(EdgePickData,Blur2Data,Width,Width,Height);
     else  EdgeDetectorFloatCanny(EdgePickData,Blur2Data,Width,Width,Height);
  }
  else
  {
    if (!Canny)  EdgeDetectorYV12(EdgePickData,Src,SrcPitch,Width,Height);
     else  EdgeDetectorYV12Canny(EdgePickData,Src,SrcPitch,Width,Height);
  }
}



void  ShockTransportYUY2(BYTE  Dest[],  const  BYTE  Src[],  signed char  EdgePickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch,  double  Intensity)
{
  if (Intensity==1.0)
  {
    PickLumaYUY2(Dest,EdgePickData,Src,EdgePickData,DestPitch,SrcPitch,Width,Height);
    PickChromaYUY2(Dest+1,Src+1,EdgePickData,DestPitch,SrcPitch,Width,Height);
    PickChromaYUY2(Dest+3,Src+3,EdgePickData,DestPitch,SrcPitch,Width,Height);
  }
  else
  {
    PickLumaYUY2Intensity(Dest,EdgePickData,Src,EdgePickData,DestPitch,SrcPitch,Width,Height,Intensity);
    PickChromaYUY2Intensity(Dest+1,Src+1,EdgePickData,DestPitch,SrcPitch,Width,Height,Intensity);
    PickChromaYUY2Intensity(Dest+3,Src+3,EdgePickData,DestPitch,SrcPitch,Width,Height,Intensity);
  }
}



void  ShockTransportYV12(BYTE  DestY[],  BYTE  DestU[],  BYTE  DestV[],  const  BYTE  SrcY[],  const  BYTE  SrcU[],  const  BYTE  SrcV[],  signed char  EdgePickData[],  int  Width,  int  Height,  int  DestPitchY,  int  DestPitchU,  int  DestPitchV,  int  SrcPitchY,  int  SrcPitchU,  int  SrcPitchV,  double  Intensity){
  if (Intensity==1.0)
  {
    PickLumaYV12(DestY,EdgePickData,SrcY,EdgePickData,DestPitchY,SrcPitchY,Width,Height);
    PickChromaYV12(DestU,SrcU,EdgePickData,DestPitchU,SrcPitchU,Width,Height);
    PickChromaYV12(DestV,SrcV,EdgePickData,DestPitchV,SrcPitchV,Width,Height);
  }
  else
  {
    PickLumaYV12Intensity(DestY,EdgePickData,SrcY,EdgePickData,DestPitchY,SrcPitchY,Width,Height,Intensity);
    PickChromaYV12Intensity(DestU,SrcU,EdgePickData,DestPitchU,SrcPitchU,Width,Height,Intensity);
    PickChromaYV12Intensity(DestV,SrcV,EdgePickData,DestPitchV,SrcPitchV,Width,Height,Intensity);
  }
}



/**************************************************************
 This part implements all colorspace dependent functions from
 AVSShockCSdep.c.
 **************************************************************/

/*****  LineConvolution  *****/
#define  AVSShock_CSdep_LineConvolution  LineConvolutionByte
#define  AVSShock_CSdep_SrcType  BYTE
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_LineConvolution  LineConvolutionFloat
#define  AVSShock_CSdep_SrcType  float
#include  "AVSShockCSdep.c"


/*****  EdgeDetector  *****/
#define  AVSShock_CSdep_EdgeDetector  EdgeDetectorYUY2
#define  AVSShock_CSdep_SrcType  BYTE
#define  AVSShock_CSdep_SrcStride  2
#define  AVSShock_CSdep_SrcCalcType  signed int
#define  AVSShock_CSdep_EdgeCanny  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_EdgeDetector  EdgeDetectorYV12
#define  AVSShock_CSdep_SrcType  BYTE
#define  AVSShock_CSdep_SrcStride  1
#define  AVSShock_CSdep_SrcCalcType  signed int
#define  AVSShock_CSdep_EdgeCanny  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_EdgeDetector  EdgeDetectorFloat
#define  AVSShock_CSdep_SrcType  float
#define  AVSShock_CSdep_SrcStride  1
#define  AVSShock_CSdep_SrcCalcType  double
#define  AVSShock_CSdep_EdgeCanny  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_EdgeDetector  EdgeDetectorYUY2Canny
#define  AVSShock_CSdep_SrcType  BYTE
#define  AVSShock_CSdep_SrcStride  2
#define  AVSShock_CSdep_SrcCalcType  signed int
#define  AVSShock_CSdep_EdgeCanny  1
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_EdgeDetector  EdgeDetectorYV12Canny
#define  AVSShock_CSdep_SrcType  BYTE
#define  AVSShock_CSdep_SrcStride  1
#define  AVSShock_CSdep_SrcCalcType  signed int
#define  AVSShock_CSdep_EdgeCanny  1
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_EdgeDetector  EdgeDetectorFloatCanny
#define  AVSShock_CSdep_SrcType  float
#define  AVSShock_CSdep_SrcStride  1
#define  AVSShock_CSdep_SrcCalcType  double
#define  AVSShock_CSdep_EdgeCanny  1
#include  "AVSShockCSdep.c"


/*****  PickLuma  *****/
#define  AVSShock_CSdep_PickLuma  PickLumaYUY2
#define  AVSShock_CSdep_DestStride  2
#define  AVSShock_CSdep_SrcStride  2
#define  AVSShock_CSdep_PickStore  1
#define  AVSShock_CSdep_PickEroDil9  ErosionDilation3x3
#define  AVSShock_CSdep_Intensity  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_PickLuma  PickLumaYV12
#define  AVSShock_CSdep_DestStride  1
#define  AVSShock_CSdep_SrcStride  1
#define  AVSShock_CSdep_PickStore  1
#define  AVSShock_CSdep_PickEroDil9  ErosionDilation3x3
#define  AVSShock_CSdep_Intensity  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_PickLuma  PickLumaYUY2Intensity
#define  AVSShock_CSdep_DestStride  2
#define  AVSShock_CSdep_SrcStride  2
#define  AVSShock_CSdep_PickStore  1
#define  AVSShock_CSdep_PickEroDil9  ErosionDilation3x3
#define  AVSShock_CSdep_Intensity  1
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_PickLuma  PickLumaYV12Intensity
#define  AVSShock_CSdep_DestStride  1
#define  AVSShock_CSdep_SrcStride  1
#define  AVSShock_CSdep_PickStore  1
#define  AVSShock_CSdep_PickEroDil9  ErosionDilation3x3
#define  AVSShock_CSdep_Intensity  1
#include  "AVSShockCSdep.c"


/*****  PickChroma  *****/
#define  AVSShock_CSdep_PickChroma  PickChromaYUY2
#define  AVSShock_CSdep_ChromaStride  4
#define  AVSShock_CSdep_ChromaWidth  2
#define  AVSShock_CSdep_ChromaHeight  1
#define  AVSShock_CSdep_Intensity  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_PickChroma  PickChromaYV12
#define  AVSShock_CSdep_ChromaStride  1
#define  AVSShock_CSdep_ChromaWidth  2
#define  AVSShock_CSdep_ChromaHeight  2
#define  AVSShock_CSdep_Intensity  0
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_PickChroma  PickChromaYUY2Intensity
#define  AVSShock_CSdep_ChromaStride  4
#define  AVSShock_CSdep_ChromaWidth  2
#define  AVSShock_CSdep_ChromaHeight  1
#define  AVSShock_CSdep_Intensity  1
#include  "AVSShockCSdep.c"

#define  AVSShock_CSdep_PickChroma  PickChromaYV12Intensity
#define  AVSShock_CSdep_ChromaStride  1
#define  AVSShock_CSdep_ChromaWidth  2
#define  AVSShock_CSdep_ChromaHeight  2
#define  AVSShock_CSdep_Intensity  1
#include  "AVSShockCSdep.c"
/**************************************************************/
