/**************************************************************
 This file is part of
          Shock Filter for AviSynth by Wolfgang Boiger
 Copyright (C)  2006-2008  Wolfgang Boiger, Berlin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2, as published by the Free Software Foundation in
 June 1991 (see 'GNUGPLv2.txt').
 
 There is no guaranty that the rights of third parties
 (including software patents) are not infringed upon by the
 execution or distribution of this program.
 Referring to section 8 of the General Public License, any use
 of this program (including all activities mentioned in
 section 0 and including execution) is restricted to countries
 where any of these activities do not infringe upon any such
 rights of third parties. It lies with the user to verify the
 compliance of his or her use (especially concerning but not
 limited to distribution) of this program with the rights of
 any third parties in the respective countries or regions.
 
 This program is distributed in the hope that it will be
 useful, but WITHOUT ANY WARRANTY; without even the implied
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public
 License 2 along with this program; if not, write to the
   Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **************************************************************/



/**************************************************************
 This file holds all functions which have to be implemented
 multiple times since they depend on the colorspace (and some
 other things).
 **************************************************************/



/**************************************************************
                        LineConvolution
 AVSShock_CSdep_LineConvolution
 AVSShock_CSdep_SrcType
 **************************************************************/

#ifdef  AVSShock_CSdep_LineConvolution
void  AVSShock_CSdep_LineConvolution(float  DestData[],  const  AVSShock_CSdep_SrcType  SrcData[],  int  DestStride,  int  SrcStride,  signed int  SrcIndex[],  int  Length,  double  Kernel[],  int  KernelSize)
/* Kernel should point to the *center* of the kernel, not the
   first vector index! It will be addressed like
          Kernel[-KernelSize] up to Kernel[+KernelSize]
   Index should help in addressing point outside of the SrcData,
   like SrcData[0-KernelSize] will be addressed as
     SrcData[Index[0-KernelSize]]                               */
{
  register  int  j, k;
  register  double  Data;


  for (k=0 ; k<Length ; k++)
  {
    Data = 0;
    for (j=-KernelSize ; j<=+KernelSize ; j++)  Data += 1.0*Kernel[j]*(double)SrcData[SrcStride*SrcIndex[k+j]];
    DestData[DestStride*k] = Data;
  }
}

#undef  AVSShock_CSdep_LineConvolution
#undef  AVSShock_CSdep_SrcType
#endif



/**************************************************************
                          EdgeDetector
 AVSShock_CSdep_EdgeDetector
 AVSShock_CSdep_SrcType
 AVSShock_CSdep_SrcStride
 AVSShock_CSdep_SrcCalcType
 AVSShock_CSdep_EdgeCanny  0:Marr-Hildreth  1:Haralick-Canny
 **************************************************************/

#ifdef  AVSShock_CSdep_EdgeDetector
void  AVSShock_CSdep_EdgeDetector(signed char  EdgeSign[],  const  AVSShock_CSdep_SrcType  SrcLuma[],  int  SrcPitch,  int  Width,  int  Height)
{
  int x, y;
  AVSShock_CSdep_SrcCalcType  Laplace;
  signed int  xm, xp, ym, yp;
  #if (defined(AVSShock_CSdep_EdgeCanny) && (AVSShock_CSdep_EdgeCanny!=0))
  AVSShock_CSdep_SrcCalcType  ux, uy, uxx, uxy, uyy;
  #endif


  ym = 0;  /* ym and yp store offsets for previous/next pixel line (0 at borders => image appears mirrored) */
  yp = SrcPitch;
  for (y=0 ; y<Height ; y++)
  {
    if (y==1)  ym = -SrcPitch;
    if (y==Height-1)  yp = 0;
    xm = 0;
    xp = AVSShock_CSdep_SrcStride;
    for (x=0 ; x<Width ; x++)
    {
      if (x==1)  xm = -AVSShock_CSdep_SrcStride;
      if (x==Width-1)  xp = 0;
      #if (defined(AVSShock_CSdep_EdgeCanny) && (AVSShock_CSdep_EdgeCanny!=0))
      ux         =                                SrcLuma[ym+xp];  /* Implement u''.(u',u') with Sobel-masks */
      ux        +=  2*(AVSShock_CSdep_SrcCalcType)SrcLuma[   xp];
      ux        +=                                SrcLuma[yp+xp];
      ux        -=                                SrcLuma[ym+xm];
      ux        -=  2*(AVSShock_CSdep_SrcCalcType)SrcLuma[   xm];
      ux        -=                                SrcLuma[yp+xm];
      uy         =                                SrcLuma[xm+yp];
      uy        +=  2*(AVSShock_CSdep_SrcCalcType)SrcLuma[   yp];
      uy        +=                                SrcLuma[xp+yp];
      uy        -=                                SrcLuma[xm+ym];
      uy        -=  2*(AVSShock_CSdep_SrcCalcType)SrcLuma[   ym];
      uy        -=                                SrcLuma[xp+ym];
      uyy = uxx  = -2*(AVSShock_CSdep_SrcCalcType)SrcLuma[  0  ];
      uxx       +=                                SrcLuma[ xm  ];
      uxx       +=                                SrcLuma[ xp  ];
      uyy       +=                                SrcLuma[ ym  ];
      uyy       +=                                SrcLuma[ yp  ];
      uxy        =                                SrcLuma[xp+yp];
      uxy       +=                                SrcLuma[xm+ym];
      uxy       -=                                SrcLuma[xp+ym];
      uxy       -=                                SrcLuma[xm+yp];
      uxx *= 2;
      uyy *= 2;
      Laplace = uxx*ux*ux+uxy*ux*uy+uyy*uy*uy;
      #else
      Laplace = -4*(AVSShock_CSdep_SrcCalcType)SrcLuma[0];  /* Calculate Laplace of smoothed image */
      Laplace +=                               SrcLuma[xm];
      Laplace +=                               SrcLuma[xp];
      Laplace +=                               SrcLuma[ym];
      Laplace +=                               SrcLuma[yp];
      #endif
      EdgeSign[0] = (Laplace>0)?(+1):((Laplace==0)?(0):(-1));
      EdgeSign++;
      SrcLuma += AVSShock_CSdep_SrcStride;
    }
    SrcLuma += SrcPitch-AVSShock_CSdep_SrcStride*Width;
  }
}
#undef  AVSShock_CSdep_EdgeDetector
#undef  AVSShock_CSdep_SrcType
#undef  AVSShock_CSdep_SrcStride
#undef  AVSShock_CSdep_SrcCalcType
#undef  AVSShock_CSdep_EdgeCanny
#endif



/**************************************************************
                           PickLuma
 AVSShock_CSdep_PickLuma
 AVSShock_CSdep_DestStride
 AVSShock_CSdep_SrcStride
 AVSShock_CSdep_PickStore
 AVSShock_CSdep_PickEroDil9
 AVSShock_CSdep_Intensity
 **************************************************************/

#ifdef  AVSShock_CSdep_PickLuma
void  AVSShock_CSdep_PickLuma(
  BYTE  DestLuma[],
  #if (defined(AVSShock_CSdep_PickStore) && (AVSShock_CSdep_PickStore!=0))
  signed char  DestPick[],
  #endif
  const  BYTE  SrcLuma[],
  signed char  EdgeSign[],
  int  DestPitch,
  int  SrcPitch,
  int  Width,
  int  Height
  #if (defined(AVSShock_CSdep_Intensity) && (AVSShock_CSdep_Intensity!=0))
  ,  double  Intensity
  #endif
  )
/* EdgeSign and DestPick may be the same arrays !! */
{
  register  int x, y;
  signed int  xm, xp, ym, yp, P;
  BYTE  Pixel[11];


  ym = 0;  /* ym and yp store offsets for previous/next pixel line (0 at borders => image appears mirrored) */
  yp = SrcPitch;
  for (y=0 ; y<Height ; y++)
  {
    if (y==1)  ym = -SrcPitch;
    if (y==Height-1)  yp = 0;
    xm = 0;
    xp = AVSShock_CSdep_SrcStride;
    for (x=0 ; x<Width ; x++)
    {
      if (x==1)  xm = -AVSShock_CSdep_SrcStride;
      if (x==Width-1)  xp = 0;
      
      if (EdgeSign[0]==0)
      {
        #if (defined(AVSShock_CSdep_PickStore) && (AVSShock_CSdep_PickStore!=0))
        DestPick[0] = 4;
        #endif
        DestLuma[0] = SrcLuma[0];
      }
      else
      {
        /*****************
          Pixel numbers:
            0    1    2
            4    5    6
            8    9   10
         *****************/
        #if (defined(AVSShock_CSdep_PickEroDil9) && (AVSShock_CSdep_PickEroDil9!=0))
        Pixel[0]  = SrcLuma[ym+xm];
        Pixel[2]  = SrcLuma[ym+xp];
        Pixel[8]  = SrcLuma[yp+xm];
        Pixel[10] = SrcLuma[yp+xp];
        #endif
        Pixel[1]  = SrcLuma[ym];
        Pixel[4]  = SrcLuma[   xm];
        Pixel[5]  = SrcLuma[  0  ];
        Pixel[6]  = SrcLuma[   xp];
        Pixel[9]  = SrcLuma[yp   ];
        P = 5;
        if (EdgeSign[0]>0)
        {
          /* Erosion */
          if (Pixel[P]>Pixel[1])   P = 1;
          if (Pixel[P]>Pixel[4])   P = 4;
          if (Pixel[P]>Pixel[6])   P = 6;
          if (Pixel[P]>Pixel[9])   P = 9;
          #if (defined(AVSShock_CSdep_PickEroDil9) && (AVSShock_CSdep_PickEroDil9!=0))
          if (Pixel[P]>Pixel[0])   P = 0;
          if (Pixel[P]>Pixel[2])   P = 2;
          if (Pixel[P]>Pixel[8])   P = 8;
          if (Pixel[P]>Pixel[10])  P = 10;
          #endif
        }
        else
        {
          /* Dilation */
          if (Pixel[P]<Pixel[1])   P = 1;
          if (Pixel[P]<Pixel[4])   P = 4;
          if (Pixel[P]<Pixel[6])   P = 6;
          if (Pixel[P]<Pixel[9])   P = 9;
          #if (defined(AVSShock_CSdep_PickEroDil9) && (AVSShock_CSdep_PickEroDil9!=0))
          if (Pixel[P]<Pixel[0])   P = 0;
          if (Pixel[P]<Pixel[2])   P = 2;
          if (Pixel[P]<Pixel[8])   P = 8;
          if (Pixel[P]<Pixel[10])  P = 10;
          #endif
        }
        #if (defined(AVSShock_CSdep_PickStore) && (AVSShock_CSdep_PickStore!=0))
        DestPick[0] = P;
        #endif
        #if (defined(AVSShock_CSdep_Intensity) && (AVSShock_CSdep_Intensity!=0))
        DestLuma[0] = (P==5)?Pixel[5]:round(Intensity*Pixel[P]+(1.0-Intensity)*Pixel[5]);
        #else
        DestLuma[0] = Pixel[P];
        #endif
      }
      #if (defined(AVSShock_CSdep_PickStore) && (AVSShock_CSdep_PickStore!=0))
      DestPick++;
      #endif
      DestLuma += AVSShock_CSdep_DestStride;
      SrcLuma += AVSShock_CSdep_SrcStride;
      EdgeSign++;
    }
    DestLuma += DestPitch-AVSShock_CSdep_DestStride*Width;
    SrcLuma += SrcPitch-AVSShock_CSdep_SrcStride*Width;
  }
}
#undef  AVSShock_CSdep_PickLuma
#undef  AVSShock_CSdep_DestStride
#undef  AVSShock_CSdep_SrcStride
#undef  AVSShock_CSdep_PickStore
#undef  AVSShock_CSdep_PickEroDil9
#undef  AVSShock_CSdep_Intensity
#endif



/**************************************************************
                          PickChroma
 AVSShock_CSdep_PickChroma
 AVSShock_CSdep_ChromaStride
 AVSShock_CSdep_ChromaWidth
 AVSShock_CSdep_ChromaHeight
 AVSShock_CSdep_Intensity
 **************************************************************/

#ifdef  AVSShock_CSdep_PickChroma
void  AVSShock_CSdep_PickChroma(
  BYTE  DestChroma[],
  const  BYTE  SrcChroma[],
  signed char  Pick[],
  int  DestPitch,
  int  SrcPitch,
  int  Width,
  int  Height
  #if (defined(AVSShock_CSdep_Intensity) && (AVSShock_CSdep_Intensity!=0))
  ,  double  Intensity
  #endif
  )
{
  #if (AVSShock_CSdep_ChromaHeight==1)
  #define n 0
  #else
  int  n;
  #endif
  #if (AVSShock_CSdep_ChromaWidth==1)
  #define m 0
  #else
  int  m;
  #endif
  
  int  x, y;
  signed int  xc[3], yc[3];  /* Offsets of current position to address neighbor pixels */
  long  Value;
  signed char  P;


  xc[1] = yc[1] = 0;
  #if (AVSShock_CSdep_ChromaHeight==1)
  yc[0] = 0;
  yc[2] = SrcPitch;
  #endif
  for (y=0 ; y<Height/AVSShock_CSdep_ChromaHeight ; y++)
  {
    #if (AVSShock_CSdep_ChromaHeight==1)
    if (y==1)  yc[0] = -SrcPitch;
    if (y==Height-1)  yc[2] = 0;
    #endif
    #if (AVSShock_CSdep_ChromaWidth==1)
    xc[0] = 0;
    xc[2] = AVSShock_CSdep_ChromaStride;
    #endif
    for (x=0 ; x<Width/AVSShock_CSdep_ChromaWidth ; x++)
    {
      #if (AVSShock_CSdep_ChromaWidth==1)
      if (x==1)  xc[0] = -AVSShock_CSdep_ChromaStride;
      if (x==Width-1)  xc[2] = 0;
      #endif
      #if ((AVSShock_CSdep_ChromaHeight>2)&&(AVSShock_CSdep_ChromaWidth>2))
      Value = (AVSShock_CSdep_ChromaHeight-2)*(AVSShock_CSdep_ChromaWidth-2)*SrcChroma[AVSShock_CSdep_ChromaStride*x];
      #elif (!((AVSShock_CSdep_ChromaHeight==1)&&(AVSShock_CSdep_ChromaWidth==1)))
      Value = 0;
      #endif
      #if (AVSShock_CSdep_ChromaHeight!=1)
      for (n=0 ; n<AVSShock_CSdep_ChromaHeight ; n++)
      {
        yc[0] = ((n==0) && (y>0))?(-SrcPitch):0;
        yc[2] = ((n==AVSShock_CSdep_ChromaHeight-1) && (y<Height/AVSShock_CSdep_ChromaHeight-1))?(+SrcPitch):0;
      #else
      {
      #endif
        #if (AVSShock_CSdep_ChromaWidth!=1)
        for (m=0 ; m<AVSShock_CSdep_ChromaWidth ; m++)
        {
          xc[0] = ((m==0) && (x>0))?(-AVSShock_CSdep_ChromaStride):0;
          xc[2] = ((m==AVSShock_CSdep_ChromaWidth-1) && (x<Width/AVSShock_CSdep_ChromaWidth-1))?(+AVSShock_CSdep_ChromaStride):0;
        #else
        {
        #endif
          #if ((AVSShock_CSdep_ChromaHeight>2)&&(AVSShock_CSdep_ChromaWidth>2))
          if ((n!=0) && (m!=0) && (n!=AVSShock_CSdep_ChromaHeight-1) && (m!=AVSShock_CSdep_ChromaWidth-1))  continue;
          #endif
          P = Pick[m+Width*n];
          /*****************
            Pixel numbers:
              0    1    2
              4    5    6
              8    9   10
           *****************/
          #if ((AVSShock_CSdep_ChromaHeight==1)&&(AVSShock_CSdep_ChromaWidth==1))
          Value = SrcChroma[yc[P/4]+xc[P%4]];
          #else
          Value += SrcChroma[yc[P/4]+xc[P%4]];
          #endif
        }
      }
      #if (defined(AVSShock_CSdep_Intensity) && (AVSShock_CSdep_Intensity!=0))
      DestChroma[0] = round(Intensity*Value/AVSShock_CSdep_ChromaHeight/AVSShock_CSdep_ChromaWidth+(1.0-Intensity)*SrcChroma[0]);
      #elif ((AVSShock_CSdep_ChromaHeight==1)&&(AVSShock_CSdep_ChromaWidth==1))
      DestChroma[0] = Value;
      #else
      DestChroma[0] = round(1.0*Value/AVSShock_CSdep_ChromaHeight/AVSShock_CSdep_ChromaWidth);
      #endif
      DestChroma += AVSShock_CSdep_ChromaStride;
      SrcChroma += AVSShock_CSdep_ChromaStride;
      Pick += AVSShock_CSdep_ChromaWidth;
    }
    DestChroma += DestPitch-(Width*AVSShock_CSdep_ChromaStride)/AVSShock_CSdep_ChromaWidth;
    SrcChroma += SrcPitch-(Width*AVSShock_CSdep_ChromaStride)/AVSShock_CSdep_ChromaWidth;
  }
  #ifdef n
  #undef n
  #endif
  #ifdef m
  #undef m
  #endif
}
#undef  AVSShock_CSdep_PickChroma
#undef  AVSShock_CSdep_ChromaStride
#undef  AVSShock_CSdep_ChromaWidth
#undef  AVSShock_CSdep_ChromaHeight
#undef  AVSShock_CSdep_Intensity
#endif
